import os
from pathlib import Path

import pkg_resources
from pkg_resources import DistInfoDistribution, VersionConflict
from loguru import logger
from .BaseCommand import BaseCommand


class ValidateRequirementsCommand(BaseCommand):

    _REQUIREMENTS_PATH = os.getcwd().join('../requirements.txt')

    def execute(self):
        logger.info("VERIFYING REQUIREMENTS")

        requirements = pkg_resources.parse_requirements(ValidateRequirementsCommand._REQUIREMENTS_PATH.open())
        for requirement in requirements:
            requirement = str(requirement)
            with self.subTest(requirement=requirement):
                pkg_resources.require(requirement)



def get_sql_sentence(account_id, last_record, block):
    return '''
            SELECT
                {account_id} as account,
                False as indexed,
                
                CONCAT(DATE_FORMAT(statistics.datetime, '%Y-%m-%dT%T'), '+00:00') as tr_datetime,
                statistics.`value` as tr_value,
                statistics.type as tr_type,
                branches.branchname as tr_library,
                
                items.barcode as item_barcode,
                biblio.title as item_title,
                biblio.author as item_author,
                itemtypes.description as item_type,
                biblioitems.isbn as item_isbn,
                biblioitems.publishercode as item_publisher_code,
                biblio.biblionumber as item_biblio_number,
                
                borrowers.cardnumber as user_cardnumber,
                borrowers.userid as user_id,
                borrowers.email as user_email,
                categories.description as user_category
            FROM
                branches
            INNER JOIN statistics ON statistics.branch = branches.branchcode
            INNER JOIN items ON statistics.itemnumber = items.itemnumber
            INNER JOIN biblio ON items.biblionumber = biblio.biblionumber
            INNER JOIN itemtypes ON statistics.itemtype = itemtypes.itemtype
            INNER JOIN biblioitems ON biblio.biblionumber = biblioitems.biblionumber
            INNER JOIN borrowers ON statistics.borrowernumber = borrowers.borrowernumber
            INNER JOIN categories ON borrowers.categorycode = categories.categorycode
            
            WHERE
                statistics.datetime > "{last_record}"
            
            ORDER BY statistics.datetime ASC
            
            LIMIT {block}
            '''.format(account_id=account_id, last_record=last_record, block=block)



import sys
from loguru import logger

from src.core.DBPool import DBPool
from src.commands.BaseCommand import BaseCommand


class ValidateKohaDBCommand(BaseCommand):

    def __init__(self):
        pass

    def execute(self):
        logger.info("VALIDATING KOHA DB CONNECTION")

        if not DBPool.get_connection(insist=False):
            sys.exit(0)

        logger.success("Connection established successfully")



from loguru import logger

from src.commands.SyncKohaRowsCommand import SyncKohaRowsCommand
from src.commands.LoadConfigFileCommand import LoadConfigFileCommand
from src.commands.ValidateApiCommand import ValidateApiCommand
from src.commands.ValidateKohaDBCommand import ValidateKohaDBCommand

import click


@click.command()
@click.option('-c', '--config-file', 'config_file', required=True, help="path to configuration file")
@click.option('-v', '--verbose', 'verbosity', count=True, help="activate verbosity output")
@click.option('-r', '--run-once', 'run_once', is_flag=True, help="synchronize once and stops execution")
@click.option('-t', '--test', 'test', is_flag=True, help="run validations and stops before synchronize")
def main(config_file, verbosity, run_once, test):
    # click.secho('Hello {0} {1} {3} {2}'.format(config_file, verbosity, test, run_once))
    # click.secho(f'Hello {config_file} {verbosity} {run_once} {test}')

    # TODO
    # verify requirements before execution
    # ValidateRequirementsCommand().execute()
    # print out the current version

    # verify and load configuration file
    LoadConfigFileCommand(file_path=config_file).execute()

    # validate API availability
    ValidateApiCommand().execute()

    # validate Koha DB availability
    ValidateKohaDBCommand().execute()

    # synchronize koha registries
    while True:

        if test:
            logger.info("STOPPING EXCECUTION BECAUSE OF TEST FLAG !")
            break

        SyncKohaRowsCommand().execute()

        if run_once:
            logger.info("STOPPING EXCECUTION BECAUSE OF RUN ONCE FLAG !")
            break


# @click.command()
# @click.option('--count', default=1, help='number of greetings')
# @click.argument('name')
# def hello(count, name):

# @click.option('--count', default=1, help='number of greetings')
# @click.argument('name')
# @click.command()
# def init_db():
#     click.echo('initialized the db')

# @cli.command()
# def perform_query():
#     click.echo('querying ...')

# @click.command()
# def drop_db():
#     click.echo('dropped the db')

# cli.add_command(init_db)
# cli.add_command(drop_db)
# cli.add_command(hello)



from .Singleton import Singleton


class Logger(Singleton):

    def __new__(cls, *args, **kwargs):
        return super(Logger, cls).__new__(cls, *args, **kwargs)


# logger.info('Verifying config file')
# logger.success('Verifying config file')
# logger.warning('Verifying config file')
# logger.error('Verifying config file')
# logger.critical('Verifying config file')



import pymysql
from loguru import logger

from src.core.Config import Config

import time


class DBPool:

    @staticmethod
    def get_connection(insist=False):

        while True:
            try:
                db = pymysql.connect(db=Config.config['db']['schema'], user=Config.config['db']['user'],
                                     password=Config.config['db']['password'], host=Config.config['db']['host'],
                                     port=Config.config['db']['port'], cursorclass=pymysql.cursors.DictCursor)

                logger.success(f"Getting new connection from pool : insist -> {insist}")
                return db
            except Exception as e:
                fail = True
                logger.error(f"Koha Database connection failed : {e.args}")
                time.sleep(5)

            if not insist:
                if fail:
                    return None
                break



class Singleton(object):

    _instance = None

    def __new__(cls, *args, **kwargs):
        """"
        called before __init__
        """
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(cls, *args, **kwargs)
            return cls._instance



from loguru import logger
from tqdm import tqdm

import pprint
import time
import requests
import sys

from src.commands.BaseCommand import BaseCommand
from src.commands.LoadConfigFileCommand import LoadConfigFileCommand
from src.core import SqlSentences
from src.core.DBPool import DBPool
from src.core.Config import Config


class SyncKohaRowsCommand(BaseCommand):

    def __init__(self):
        self.add = 'api/koha'

    def execute(self):
        logger.info("STARTING SYNCHRONIZATION LOOP")

        try:
            with DBPool.get_connection(True).cursor() as cursor:

                last_record = LoadConfigFileCommand.get_last_record()
                cursor.execute(SqlSentences.get_sql_sentence(Config.config["target"]["account_id"], last_record,
                                                             Config.config["target"]["block"]))
                records = cursor.fetchall()

                if not records:
                    logger.info("There aren't new records to synchronize. Bypassing excecution")
                    time.sleep(10)
                    return

                with tqdm(total=len(records), desc="SYNCHRONIZING ROWS") as progress_bar:
                    for record in records:
                        progress_bar.update(100 / len(records))
                        for key in record:
                            if key != "tr_datetime":
                                record[key] = SyncKohaRowsCommand.sanitize(record[key])

                post_rows_req = requests.post(f"{Config.config['target']['url_add']}/{self.add}",
                                              json={
                                                  "llave": Config.config['target']['key'],
                                                  "secreto": Config.config['target']['secret'],
                                                  "docs": records
                                              })

                if post_rows_req.status_code is not 200:
                    logger.warning(f"Request status was not 200 : {post_rows_req}")

                # write last synchronized record date without timezone
                LoadConfigFileCommand.set_last_record(records[-1]['tr_datetime'].split("+")[0])
                logger.success(f"Koha records synchronized {post_rows_req.json()}")

        except Exception as e:
            logger.exception(e)
            sys.exit(logger.error(f"There was an error while synchronizing records : {e.args}"))

    """ lower case and remove noise characters """
    @staticmethod
    def sanitize(value):
        chars_to_remove = [';', ':', '/', '.', ',', '(', ')']
        if isinstance(value, str):
            value = value.lower()
            value = value.translate({ord(char): '' for char in chars_to_remove})
            value = ' '.join(value.split())
        return value



import pprint
from datetime import datetime

from loguru import logger
from jsonschema import validate

from src.commands.BaseCommand import BaseCommand
from src.core.Config import Config

import os
import sys
import yaml


class LoadConfigFileCommand(BaseCommand):

    yml_schema = """
    type: object
    properties:
      testing:
        type: array
        items:
          enum:
            - this
            - is
            - a
            - test
    """

    def __init__(self, **kwargs):
        self.file_path = kwargs.get('file_path')
        self.file_content = None

    def execute(self):
        logger.info("VERIFYING AND LOADING CONFIG FILE")

        if not os.path.isfile(self.file_path):
            sys.exit(logger.error(f"Configuration file {self.file_path} does not exist"))
        else:
            logger.success("Configuration file exists")

            try:
                with open(self.file_path, mode='r+', encoding='utf-8') as config_file:
                    self.file_content = config_file.read()
                    logger.success("Configuration file is readable and writable")

            except IOError as e:
                sys.exit(logger.error(f"Configuration file {self.file_path} is not readable or writable : {e.args}"))

            try:
                self.file_content = yaml.load(self.file_content, Loader=yaml.FullLoader)
                validate(self.file_content, yaml.load(LoadConfigFileCommand.yml_schema, Loader=yaml.FullLoader))
                logger.success("Configuration file has valid syntax")

            except yaml.YAMLError as e:
                sys.exit(logger.error(f"Configuration file {self.file_path} has format errors : {e.args}"))

            if "db" not in self.file_content:
                sys.exit(logger.error("Configuration file has not : 'db"))
            else:
                logger.success("Configuration file has : 'db'")

                for key, value in Config.config['db'].items():

                    if key not in self.file_content['db']:
                        sys.exit(logger.error(f"Configuration file missing required attribute : [db][{key}]"))

                    else:
                        logger.success(f"Configuration file has required attribute : [db][{key}]")
                        if self.file_content['db'][key] is None:
                            sys.exit(logger.error(f"Configuration file has not value : [db][{key}]"))
                        else:
                            Config.config['db'][key] = self.file_content['db'][key]
                            logger.success(f"Configuration file has required value : [db][{key}]")

            if "target" not in self.file_content:
                sys.exit(logger.error("Configuration file has not : 'target"))
            else:
                logger.success("Configuration file has : 'target'")

                for key, value in Config.config['target'].items():

                    if key not in self.file_content['target']:
                        sys.exit(logger.error(f"Configuration file missing required attribute : [target][{key}]"))

                    else:
                        logger.success(f"Configuration file has required attribute : [target][{key}]")

                        if self.file_content['target'][key] is None:
                            sys.exit(logger.error(f"Configuration file has not value : [target][{key}]"))
                        else:
                            Config.config['target'][key] = self.file_content['target'][key]
                            logger.success(f"Configuration file has required value : [target][{key}] = "
                                           f"{self.file_content['target'][key]}")

        Config.path = self.file_path

    @staticmethod
    def get_last_record():
        try:
            if not LoadConfigFileCommand.is_state():
                logger.warning("CREATING STATE FIELD : There is not previous state")

                with open(Config.path, mode='a', encoding='utf-8') as config_file:
                    state = {
                        "last_record_date": datetime(1970, 1, 1),
                        "accessed_count": 0
                    }
                    yaml.dump({"state": state}, config_file)
                    return state["last_record_date"]

            else:
                with open(Config.path, mode='r', encoding='utf-8') as config_file:
                    return yaml.load(config_file, Loader=yaml.FullLoader)['state']['last_record_date']

        except IOError as e:
            sys.exit(logger.error(f"There was a problem while reading state : {e.args}"))

    @staticmethod
    def set_last_record(date):
        try:
            if LoadConfigFileCommand.is_state():

                with open(Config.path, mode='r', encoding='utf-8') as config_file:
                    config_file_content = yaml.load(config_file, Loader=yaml.FullLoader)
                    config_file_content['state']['last_record_date'] = date
                    config_file_content['state']['accessed_count'] = config_file_content['state']['accessed_count'] + 1

                with open(Config.path, mode='w', encoding='utf-8') as config_file:
                    logger.warning(f"UPDATING STATE FIELD : [last_record_date] {date}")
                    yaml.dump(config_file_content, config_file)

        except IOError as e:
            sys.exit(logger.error(f"There was a problem while setting state : {e.args}"))

    @staticmethod
    def is_state() -> bool:
        try:
            with open(Config.path, mode='r', encoding='utf-8') as config_file:
                config_file_content = yaml.load(config_file, Loader=yaml.FullLoader)
                if 'state' not in config_file_content:
                    return False
            return True
        except IOError as e:
            sys.exit(logger.error(f"There was a problem while validating state : {e.args}"))





from setuptools import setup, find_packages

version = '1.1.1'

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name='koha-sync',
    version=version,
    license='MIT',
    author='Juan Pablo Aguirre',
    author_email='jaguirre@referencistas.com',
    description='This projects allows to synchronize remote Koha rows to an endpoint',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://jaguirre_referencistas@bitbucket.org/jaguirre_referencistas/koha_synchronizer.git',
    packages=find_packages(),
    entry_points={
      "console_scripts": ['koha-sync = src.kohaSync:main']
    },
    python_requires='>=2.7',
    install_requires=[
        'click',
        'tqdm',
        'jsonschema',
        'requests',
        'pymysql',
        'loguru',
        'setuptools',
        'wheel'
    ]
)


# execute build
#   > pip install --user --upgrade twine
#   > python3 setup.py sdist bdist_wheel
#       - tar.gz = source archive
#       - whl    = built distribution

# upload package
#   > pip install --user --upgrade twine
#   > twine upload --repository pypi dist/*

# install
#   > install python3.7 and pip3
#   > install setuptools wheel
#   > alias python="python3.7"
#   > apt-get install python3-pip
#   > pip3 install virtualenv
#   > virtualenv venv
#   > source venv/bin/activate | deactivate
#   > pip3 install --index-url https://pypi.org/ --no-deps koha-sync




class Config:

    path = None

    config = {
        "db": {
            "host": None,
            "user": None,
            "password": None,
            "port": None,
            "schema": None
        },
        "target": {
            "account_id": None,
            "url_status": None,
            "url_add": None,
            "key": None,
            "secret": None,
            "block": None,
        }
    }

    def get_instance(self):
        pass



from loguru import logger

from src.commands.BaseCommand import BaseCommand
from src.core.Config import Config

import requests
import sys


class ValidateApiCommand(BaseCommand):

    def __init__(self):
        self.status = 'status'
        self.headers = {
            "Accept": "application/json"
        }

    def execute(self):

        logger.info("VALIDATING API AVAILABILITY")

        try:
            status_req = requests.get(f"{Config.config['target']['url_status']}/{self.status}", headers=self.headers)
        except Exception as e:
            sys.exit(logger.error(f"There was an error for request  : {e.args}"))

        if status_req.status_code is not 200:
            sys.exit(logger.error(f"API {Config.config['target']['url_status']}/{self.status} "
                                  f"status code {status_req.status_code}"))

        else:
            logger.success(f"API status code : 200")

            if status_req.json()['status'] != "ok":
                sys.exit(logger.error(f"API not available"))
            else:
                logger.success(f"API availability : OK")

        # TODO
        # validate /api/koha API availability
